<?php
/**
 * @category   Mad
 * @package    Mad_Model
 * @subpackage ConnectionAdapter
 * @copyright  (c) 2007-2008 Maintainable Software, LLC
 * @license    http://opensource.org/licenses/bsd-license.php BSD
 */

/**
 * @category   Mad
 * @package    Mad_Model
 * @subpackage ConnectionAdapter
 * @copyright  (c) 2007-2008 Maintainable Software, LLC
 * @license    http://opensource.org/licenses/bsd-license.php BSD
 */
class Mad_Model_ConnectionAdapter_Abstract_TableDefinition implements ArrayAccess
{
    protected $_name    = null;
    protected $_base    = null;
    protected $_options = null;
    protected $_columns = null;

    /**
     * Class Constructor
     *
     * @param  string  $name
     * @param  string  $base
     * @param  array   $options
     */
    public function __construct($name, $base, $options=array()) 
    {
        $this->_name    = $name;
        $this->_base    = $base;
        $this->_options = $options;
        $this->_columns = array();
    }

    /**
     * @return  string
     */
    public function getName()
    {
        return $this->_name;
    }

    /**
     * @return  array
     */
    public function getOptions()
    {
        return $this->_options;
    }

    /**
     * @param   string  $name
     */
    public function primaryKey($name)
    {
        $natives = $this->_native();
        $this->column($name, $natives['primaryKey']);
    }
    
    /**
     * Instantiates a new column for the table.
     * The +type+ parameter must be one of the following values:
     * <tt>:primary_key</tt>, <tt>:string</tt>, <tt>:text</tt>,
     * <tt>:integer</tt>, <tt>:float</tt>, <tt>:datetime</tt>,
     * <tt>:timestamp</tt>, <tt>:time</tt>, <tt>:date</tt>,
     * <tt>:binary</tt>, <tt>:boolean</tt>.
     *
     * Available options are (none of these exists by default):
     * * <tt>:limit</tt>:
     *   Requests a maximum column length (<tt>:string</tt>, <tt>:text</tt>,
     *   <tt>:binary</tt> or <tt>:integer</tt> columns only)
     * * <tt>:default</tt>:
     *   The column's default value.  You cannot explicitely set the default
     *   value to +NULL+.  Simply leave off this option if you want a +NULL+
     *   default value.
     * * <tt>:null</tt>:
     *   Allows or disallows +NULL+ values in the column.  This option could
     *   have been named <tt>:null_allowed</tt>.
     *
     * This method returns <tt>self</tt>.
     *
     * ===== Examples
     *  # Assuming def is an instance of TableDefinition
     *  def.column(:granted, :boolean)
     *    #=> granted BOOLEAN
     *
     *  def.column(:picture, :binary, :limit => 2.megabytes)
     *    #=> picture BLOB(2097152)
     *
     *  def.column(:sales_stage, :string, :limit => 20, :default => 'new', :null => false)
     *    #=> sales_stage VARCHAR(20) DEFAULT 'new' NOT NULL
     * 
     * @return  TableDefinition
     */
    public function column($name, $type, $options=array())
    {
        if ($this[$name]) {
            $column = $this[$name];
        } else {
            $column = new Mad_Model_ConnectionAdapter_Abstract_ColumnDefinition(
                $this->_base, $name, $type);
        }

        $natives = $this->_native();
        $opt = $options;

        if (isset($opt['limit']) || isset($natives[$type])) {
            $nativeLimit = isset($natives[$type]['limit']) ? $natives[$type]['limit'] : null;
            $column->setLimit(isset($opt['limit']) ? $opt['limit'] : $nativeLimit);
        }

        $column->setPrecision(isset($opt['precision']) ? $opt['precision'] : null);
        $column->setScale(isset($opt['scale'])         ? $opt['scale']     : null);
        $column->setDefault(isset($opt['default'])     ? $opt['default']   : null);
        $column->setNull(isset($opt['null'])           ? $opt['null']      : null);

        $this[$name] ? $this[$name] = $column : $this->_columns[] = $column;
        return $this;
    }
    
    /**
     * Wrap up table creation block & create the table
     */
    public function end()
    {
        $conn = Mad_Model_Base::connection(); 

        // drop previous
        if (isset($this->_options['force'])) {
            $conn->dropTable($this->getName(), $this->_options);
        }
        $temp = !empty($this->_options['temporary']) ? 'TEMPORARY'                : null;
        $opts = !empty($this->_options['options'])   ? $this->_options['options'] : null;

        $sql  = "CREATE $temp TABLE ".$this->getName()." (\n".
                  $this->toSql()."\n".
                ") $opts";
        return $conn->execute($sql);
    }

    /**
     * Returns a String whose contents are the column definitions
     * concatenated together.  This string can then be pre and appended to
     * to generate the final SQL to create the table.
     * 
     * @return  string
     */
    public function toSql()
    {
        $cols = array();
        foreach ($this->_columns as $col) { $cols[] = $col->toSql(); }
        
        return "  ".implode(", \n  ", $cols);
    }


    /*##########################################################################
    # ArrayAccess
    ##########################################################################*/

    /**
     * ArrayAccess: Check if the given offset exists
     * 
     * @param   int     $offset
     * @return  boolean
     */
    public function offsetExists($offset)
    {
        foreach ($this->_columns as $column) {
            if ($column->getName() == $offset) return true;
        }
        return false;
    }

    /**
     * ArrayAccess: Return the value for the given offset.
     *
     * @param   int     $offset
     * @return  object  {@link {@baseModel}
     */
    public function offsetGet($offset)
    {
        if (!$this->offsetExists($offset)) {
            return null;
        }

        foreach ($this->_columns as $column) {
            if ($column->getName() == $offset) {
                return $column;
            }
        }
    }

    /**
     * ArrayAccess: Set value for given offset
     * 
     * @param   int     $offset
     * @param   mixed   $value
     */
    public function offsetSet($offset, $value)
    {
        foreach ($this->_columns as $key=>$column) {
            if ($column->getName() == $offset) {
                $this->_columns[$key] = $value;
            }
        }
    }

    /**
     * ArrayAccess: remove element
     * 
     * @param   int     $offset
     */
    public function offsetUnset($offset)
    {
        foreach ($this->_columns as $key=>$column) {
            if ($column->getName() == $offset) {
                unset($this->_columns[$key]);
            }
        }
    }


    /*##########################################################################
    # Protected
    ##########################################################################*/

    /**
     * Get the types
     */
    protected function _native()
    {
        return $this->_base->nativeDatabaseTypes();
    }
}